import numpy as np
from scipy import io


class Env:
    def __init__(self):
        super(Env, self).__init__()

        mat_file = io.loadmat('ZRM_Model.mat')
        self.A = mat_file['A']
        self.B = mat_file['B']
        self.C = mat_file['C']
        self.u0 = mat_file['u0']
        self.d = mat_file['d']

        self.nx = len(self.A)
        self.ny = len(self.C)
        self.nu = len(self.B[0])

        self.action_space = np.zeros([1, 9], dtype='float32')
        self.u = np.zeros([1, 9], dtype='float32')
        self.n_actions = len(self.action_space)
        self.x = np.zeros([1, self.nx], dtype='float32')
        self.y = np.zeros([1, self.ny], dtype='float32')
        self.n_states = len(self.y)
        self.steps = 1
        self.min_y = max(abs(self.d[0]))

    def reset(self):
        self.u = np.zeros([1, self.nu], dtype='float32')
        self.x = np.zeros([1, self.nx], dtype='float32')  # 초기상태로 reset
        self.y = np.zeros([1, self.ny], dtype='float32')
        self.min_y = max(abs(self.d[0]))
        self.steps = 1
        tmp_y = self.d.T

        return np.concatenate((tmp_y.T, self.u), axis=1)

    def step(self, action):
        x = self.x

        u = np.reshape(self.u + action*8, [1, 9])

        tmp = u + self.u0
        #for k in range(6):
        #    un = tmp[0, k]
        #    if (un <= 0) and (un >= 100):
        #        u[0, k] = u[0, k]

        #for k in range(7, 8):
        #    un = tmp[0, k]
        #    if (un <= 40) and (un >= 120):
        #        u[0, k] = self.u[0, k]

        self.u = u
        # action 에 따른 State Transition
        tmp_x = np.dot(self.A, x.T) + np.dot(self.B, u.T)
        self.x = tmp_x.T
        tmp_y = np.dot(self.C, self.x.T) + self.d.T
        y = tmp_y.T


        # 보상함수
        if (self.steps > 199) or (max(abs(y[0])) > 50):
            done = True
            reward = -20.0

        elif max(abs(y[0])) < 15:
            done = True
            reward = 10.0

        elif max(abs(y[0])) < self.min_y:
            done = False
            reward = self.min_y - max(abs(y[0]))
            self.min_y = max(abs(y[0]))
            self.steps += 1

        else:
            reward = 0.0
            done = False
            self.steps += 1

        self.y = y

        return np.concatenate((y, u), axis=1), reward, done
