from keras.layers import Dense, Input
#from keras.optimizers import RMSprop
from keras.optimizers import Adam
from keras import backend as K
from keras.models import Model
import tensorflow as tf
import numpy as np
import threading
import time
import pylab
from environment_a3c_v5 import Env

# 멀티쓰레딩을 위한 글로벌 변수
global episode
global episodes, y_mins, avg_ymins, avg_scores, scores, yfs, avg_yfs

episode = 0
EPISODES = 50000
# 환경 생성
env_name = "ZRM"
sv = 38

# A3CAgent 클래스(글로벌신경망)
class A3CAgent:
    def __init__(self, action_size):
        # 상태크기와 행동크기를 갖고옴
        self.state_size = (sv + 9, )
        self.action_size = action_size
        # A3C 하이퍼파라미터
        self.discount_factor = 0.99
        self.no_op_steps = 30
        self.actor_lr = 1e-4
        self.critic_lr = 1e-4
        # 쓰레드의 갯수
        self.threads = 8

        self.epsilon = 1.0
        self.epsilon_decay = 0.95
        self.epsilon_min = 0.01

        # 정책신경망과 가치신경망을 생성
        self.actor, self.critic = self.build_model()
        # 정책신경망과 가치신경망을 업데이트하는 함수 생성
        self.optimizer = [self.actor_optimizer(), self.critic_optimizer()]

        # 텐서보드 설정
        self.sess = tf.InteractiveSession()
        K.set_session(self.sess)
        self.sess.run(tf.global_variables_initializer())

        self.summary_placeholders, self.update_ops, self.summary_op = \
            self.setup_summary()
        self.summary_writer = \
            tf.summary.FileWriter('summary/zrm_a3c', self.sess.graph)

    # 쓰레드를 만들어 학습을 하는 함수
    def train(self):
        # 쓰레드 수만큼 Agent 클래스 생성
        agents = [Agent(self.action_size, self.state_size,
                        [self.actor, self.critic], self.sess,
                        self.optimizer, self.discount_factor,
                        [self.summary_op, self.summary_placeholders,
                         self.update_ops, self.summary_writer])
                  for _ in range(self.threads)]

        # 각 쓰레드 시작
        for agent in agents:
            time.sleep(1)
            agent.start()

        # 10분(600초)에 한번씩 모델을 저장
        while True:
            time.sleep(60 * 10)
            self.save_model("./save_model/zrm_a3c_25ins")

    # 정책신경망과 가치신경망을 생성
    def build_model(self):
        ins = Input(shape=self.state_size)
        fc_act = Dense(50, activation='relu')(ins)
        fc_act = Dense(50, activation='relu')(fc_act)

        fc_cri = Dense(50, activation='relu')(ins)
        fc_cri = Dense(20, activation='relu')(fc_cri)

        policy = Dense(self.action_size, activation='softmax')(fc_act)
        # policy = Dense(self.action_size, activation='sigmoid')(fc)
        value = Dense(1, activation='linear')(fc_cri)

        actor = Model(inputs=ins, outputs=policy)
        critic = Model(inputs=ins, outputs=value)

        # 가치와 정책을 예측하는 함수를 만들어냄
        actor._make_predict_function()
        critic._make_predict_function()

        actor.summary()
        critic.summary()

        return actor, critic

    # 정책신경망을 업데이트하는 함수
    def actor_optimizer(self):
        action = K.placeholder(shape=[None, self.action_size])
        advantages = K.placeholder(shape=[None, ])

        policy = self.actor.output

        # 정책 크로스 엔트로피 오류함수
        action_prob = K.sum(action * policy, axis=1)
        cross_entropy = K.log(action_prob + 1e-10) * advantages
        cross_entropy = -K.sum(cross_entropy)

        # 탐색을 지속적으로 하기 위한 엔트로피 오류
        entropy = K.sum(policy * K.log(policy + 1e-10), axis=1)
        entropy = K.sum(entropy)

        # 두 오류함수를 더해 최종 오류함수를 만듬
        loss = cross_entropy + self.epsilon * entropy

        # optimizer = RMSprop(lr=self.actor_lr, rho=0.99, epsilon=0.01)
        optimizer = Adam(lr=self.actor_lr, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0)
        updates = optimizer.get_updates(self.actor.trainable_weights, [], loss)
        train = K.function([self.actor.input, action, advantages],
                           [loss], updates=updates)
        return train

    # 가치신경망을 업데이트하는 함수
    def critic_optimizer(self):
        discounted_prediction = K.placeholder(shape=(None,))

        value = self.critic.output

        # [반환값 - 가치]의 제곱을 오류함수로 함
        loss = K.mean(K.square(discounted_prediction - value))

        # optimizer = RMSprop(lr=self.critic_lr, rho=0.99, epsilon=0.01)
        optimizer = Adam(lr=self.critic_lr, beta_1=0.9, beta_2=0.999, epsilon=1e-08, decay=0.0)
        updates = optimizer.get_updates(self.critic.trainable_weights, [], loss)
        train = K.function([self.critic.input, discounted_prediction],
                           [loss], updates=updates)
        return train

    def load_model(self, name):
        self.actor.load_weights(name + "_actor.h5")
        self.critic.load_weights(name + "_critic.h5")

    def save_model(self, name):
        self.actor.save_weights(name + "_actor.h5")
        self.critic.save_weights(name + "_critic.h5")

    # 각 에피소드 당 학습 정보를 기록
    def setup_summary(self):
        episode_total_reward = tf.Variable(0.)
        episode_avg_max_q = tf.Variable(0.)
        episode_duration = tf.Variable(0.)

        tf.summary.scalar('Total Reward/Episode', episode_total_reward)
        tf.summary.scalar('Average Max Prob/Episode', episode_avg_max_q)
        tf.summary.scalar('Duration/Episode', episode_duration)

        summary_vars = [episode_total_reward,
                        episode_avg_max_q,
                        episode_duration]

        summary_placeholders = [tf.placeholder(tf.float32)
                                for _ in range(len(summary_vars))]
        update_ops = [summary_vars[i].assign(summary_placeholders[i])
                      for i in range(len(summary_vars))]
        summary_op = tf.summary.merge_all()
        return summary_placeholders, update_ops, summary_op

    def get_action(self, history):
        history = np.float32(history)
        policy = self.local_actor.predict(history)[0]
        action_index = np.random.choice(self.action_size, 1, p=policy)[0]

        return action_index, policy


# 액터러너 클래스(쓰레드)
class Agent(threading.Thread):
    def __init__(self, action_size, state_size, model, sess,
                 optimizer, discount_factor, summary_ops):
        threading.Thread.__init__(self)

        # A3CAgent 클래스에서 상속
        self.action_size = action_size
        self.state_size = state_size
        self.actor, self.critic = model
        self.sess = sess
        self.optimizer = optimizer
        self.discount_factor = discount_factor
        [self.summary_op, self.summary_placeholders,
         self.update_ops, self.summary_writer] = summary_ops

        self.epsilon = 1.0
        self.epsilon_decay = 0.9995
        self.epsilon_min = 0.01

        # 지정된 타임스텝동안 샘플을 저장할 리스트
        self.states, self.actions, self.rewards = [], [], []

        # 로컬 모델 생성
        self.local_actor, self.local_critic = self.build_local_model()

        self.avg_p_max = 0
        self.avg_loss = 0

        # 모델 업데이트 주기
        self.t_max = 20
        self.t = 0

    def run(self):
        global episode, episodes, y_mins, avg_ymins, avg_scores, scores, yfs, avg_yfs

        # env = gym.make(env_name)
        env = Env()
        step = 0
        while episode < EPISODES:
            done = False

            score = 0
            observe = env.reset()
            next_observe = observe

            while not done:
                step += 1
                self.t += 1
                observe = next_observe
                action, policy = self.get_action(np.reshape(observe, [1, sv + 9]))

                # 선택한 행동으로 한 스텝을 실행
                next_observe, reward, done = env.step(action)

                # 각 타임스텝마다 상태 전처리

                # 정책의 최대값
                self.avg_p_max += np.amax(self.actor.predict(
                    np.float32(next_observe)))

                score += reward
                # reward = np.clip(reward, -1., 1.)

                # 샘플을 저장
                self.append_sample(observe[0], action, reward)

                # 에피소드가 끝나거나 최대 타임스텝 수에 도달하면 학습을 진행
                if self.t >= self.t_max or done:
                    self.train_model(done)
                    self.update_local_model()
                    self.t = 0

                if done:
                    # 각 에피소드 당 학습 정보를 기록
                    episode += 1
                    # print("episode:", episode, "  score:", score, "  step:",
                    #
                    ym = max(abs(env.y[0]))
                    y_mins.append(env.min_y)
                    yfs.append(ym)
                    scores.append(score)

                    if episode % 20 == 0:
                        avg_ymins.append(np.mean(y_mins))
                        episodes.append(episode)
                        avg_scores.append(np.mean(scores))
                        avg_yfs.append(np.mean(yfs))

                        y_mins = []
                        yfs = []
                        scores = []

                    if episode % 50 == 0:
                        pylab.figure(1)
                        pylab.plot(episodes, avg_ymins, 'b')
                        pylab.savefig("./save_graph/zrm_a3c_ymins.png")
                        pylab.figure(2)
                        pylab.plot(episodes, avg_scores, 'r')
                        pylab.savefig("./save_graph/zrm_a3c_scores.png")
                        pylab.figure(3)
                        pylab.plot(episodes, avg_yfs, 'b')
                        pylab.savefig("./save_graph/zrm_a3c_yfs.png")

                        print("episode:", episode, "  score:", score, "  step:", step, " y_min:", env.min_y,
                              " yf_{max}:", ym, " epsilon", self.epsilon)

                    step = 0

    # k-스텝 prediction 계산
    def discounted_prediction(self, rewards, done):
        discounted_prediction = np.zeros_like(rewards)
        running_add = 0

        if not done:
            running_add = self.critic.predict(np.reshape(
                self.states[-1], [1, sv + 9]))[0]

        for t in reversed(range(0, len(rewards))):
            running_add = running_add * self.discount_factor + rewards[t]
            discounted_prediction[t] = running_add
        return discounted_prediction

    # 정책신경망과 가치신경망을 업데이트
    def train_model(self, done):
        if self.epsilon > self.epsilon_min:
            self.epsilon *= self.epsilon_decay

        discounted_prediction = self.discounted_prediction(self.rewards, done)

        states = np.zeros((len(self.states), sv + 9))
        for i in range(len(self.states)):
            states[i] = self.states[i]

        states = np.float32(states)

        values = self.critic.predict(states)
        values = np.reshape(values, len(values))

        advantages = discounted_prediction - values

        self.optimizer[0]([states, self.actions, advantages])
        self.optimizer[1]([states, discounted_prediction])
        self.states, self.actions, self.rewards = [], [], []

    # 로컬신경망을 생성하는 함수
    def build_local_model(self):
        ins = Input(shape=self.state_size)

        fc_act = Dense(50, activation='relu')(ins)
        fc_act = Dense(50, activation='relu')(fc_act)

        fc_cri = Dense(50, activation='relu')(ins)
        fc_cri = Dense(20, activation='relu')(fc_cri)

        policy = Dense(self.action_size, activation='softmax')(fc_act)
        # policy = Dense(self.action_size, activation='sigmoid')(fc)
        value = Dense(1, activation='linear')(fc_cri)

        local_actor = Model(inputs=ins, outputs=policy)
        local_critic = Model(inputs=ins, outputs=value)

        local_actor._make_predict_function()
        local_critic._make_predict_function()

        local_actor.set_weights(self.actor.get_weights())
        local_critic.set_weights(self.critic.get_weights())

        local_actor.summary()
        local_critic.summary()

        return local_actor, local_critic

    # 로컬신경망을 글로벌신경망으로 업데이트
    def update_local_model(self):
        self.local_actor.set_weights(self.actor.get_weights())
        self.local_critic.set_weights(self.critic.get_weights())

    # 정책신경망의 출력을 받아서 확률적으로 행동을 선택
    def get_action(self, history):
        history = np.float32(history)
        policy = self.local_actor.predict(history)[0]
        action_index = np.random.choice(self.action_size, 1, p=policy)[0]

        return action_index, policy

    # 샘플을 저장
    def append_sample(self, history, action, reward):
        self.states.append(history)
        act = np.zeros(self.action_size)
        act[action] = 1
        self.actions.append(act)
        self.rewards.append(reward)

if __name__ == "__main__":
    global episodes, y_mins, avg_ymins, avg_scores, scores, yfs, avg_yfs
    episodes, y_mins, avg_ymins , avg_scores, scores, yfs, avg_yfs = [], [], [], [], [], [], []
    global_agent = A3CAgent(action_size=25)
    global_agent.train()